# -*- coding: utf-8 -*-

# Copyright 2008-2014 Richard Dymond (rjdymond@gmail.com)
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.

import cgi

from .skoolhtml import Udg as BaseUdg, HtmlWriter, FileInfo, join, MAIN_CODE_ID
from .skoolasm import AsmWriter
from .skoolmacro import parse_ints, parse_params

class MicrosphereHtmlWriter(HtmlWriter):
    def init(self):
        self.char_buf_descs = self._get_char_buf_descs()
        self.keypress_routines = self.get_dictionary('KeypressRoutines')
        self.as_descs = self.get_dictionary('AnimatoryStates')
        self.characters = self.get_dictionary('Characters')

    def init_lessons(self, min_lesson, max_lesson):
        self.min_lesson = min_lesson
        self.max_lesson = max_lesson
        self.lesson_descs = self.get_dictionary('Lessons')

    def init_characters(self, min_character, max_character):
        self.min_character = min_character
        self.max_character = max_character

    def _calculate_tap_index(self, tap_address_table, max_tap):
        min_tap = tap_address_table % 256
        self.tap_addresses = {}
        self.tap_descs = {}
        for t in range(min_tap, max_tap + 1, 2):
            address = tap_address_table + t - min_tap
            tap_address = self.snapshot[address] + 256 * self.snapshot[address + 1]
            tap_entry = self.entries.get(tap_address)
            if tap_entry:
                sep_index = tap_entry.description.index(':')
                self.tap_descs[t] = tap_entry.description[sep_index + 1:].strip()
                self.tap_addresses[t] = tap_address

    def _get_sprite_udg(self, state, attr, ref_page, udg_page):
        ref_addr = state + 256 * ref_page
        ref = self.snapshot[ref_addr]
        if ref == 0:
            # These must be lists so that the UDG can be flipped or rotated
            # in-place
            udg = [0, 0, 0, 0, 0, 0, 0, 0]
            mask = [255, 255, 255, 255, 255, 255, 255, 255]
        else:
            udg_addr = ref + 256 * udg_page
            udg = self.snapshot[udg_addr:udg_addr + 4096:512]
            mask = self.snapshot[udg_addr + 256:udg_addr + 4352:512]
        return Udg(attr, udg, mask, ref_addr=ref_addr, ref=ref, udg_page=udg_page)

    def _get_sprite_size(self, state):
        return 3, 4

    def get_chr(self, code):
        if code == 127:
            return '&#169;'
        if code == 94:
            return '&#8593;'
        if code == 96:
            return '&#163;'
        return cgi.escape(chr(code))

    def build_sprite(self, state, attr=None, udg_page=None):
        width, height = self._get_sprite_size(state)
        udg_array = []
        for row in range(height):
            udg_array.append([self.get_sprite_udg(state, attr, row, col, udg_page) for col in range(width)])
        return udg_array

    def get_font_bitmap(self, character):
        base_address = ord(character) + self.font_address + 256
        width = self.snapshot[base_address - 256]
        return self.snapshot[base_address:base_address + 256 * width:256]

    def get_text(self, words):
        columns = []
        for character in words:
            columns.append(0)
            columns.extend(self.get_font_bitmap(character))
        udg_array = []
        for col in range(len(columns)):
            col_byte = columns[col]
            udg_index = col // 8
            bit = 2**(7 - (col - 8 * udg_index))
            if udg_index == len(udg_array):
                udg_array.append([0] * 8)
            udg = udg_array[udg_index]
            for b in range(8):
                udg[7 - b] |= bit * (col_byte & 1)
                col_byte //= 2
        return udg_array

    def font(self, cwd, address, num_chars=96, scale=2, fname='font'):
        img_path = self.image_path(fname, 'FontImagePath')
        if self.need_image(img_path):
            first_char = address % 256
            udgs = self.get_text(''.join([chr(c) for c in range(first_char, first_char + num_chars)]))
            udg_array = [[Udg(56, udg) for udg in udgs]]
            width = sum(self.snapshot[address:address + num_chars]) + num_chars + 1
            crop_rect = (0, 0, width * scale, None)
            self.write_image(img_path, udg_array, crop_rect, scale)
        return self.img_element(cwd, img_path)

    def _draw_text(self, udgs, words, x, y):
        text_udgs = self.get_text(words)
        for col in range(len(text_udgs)):
            udg = udgs[y][col + x].data
            for index in range(8):
                udg[index] |= text_udgs[col][index]

    def get_skool_udgs(self, x, y, w, h, show_chars=False, show_x=0):
        skool_udgs = []
        for row in range(y, y + h):
            skool_udgs.append([])
            for col in range(x, x + w):
                skool_udgs[-1].append(self.get_skool_udg(row, col, show_chars))
        if show_chars:
            self._superimpose_sprite_udgs(skool_udgs, x, y, w, h)
        if show_x is not None and show_x > 0:
            line = [Udg(8 * (7 - skool_x % 2), [0] * 8) for skool_x in range(x, x + w)]
            for skool_x in range(x, x + w):
                if skool_x % show_x == 0:
                    self._draw_text([line], str(skool_x), skool_x - x, 0)
            skool_udgs.insert(0, line)
            skool_udgs.append(line)
        return skool_udgs

    def _superimpose_sprite_udgs(self, udg_array, x, y, w, h):
        for state, char_x, char_y in self.get_initial_states():
            x0 = char_x - x
            if x0 >= w:
                continue
            x1 = min((x0 + 3, w))
            if x1 < 0:
                continue
            x0 = max((x0, 0))
            y0 = char_y - y
            if y0 >= h:
                continue
            y1 = min((y0 + 4, h))
            if y1 < 0:
                continue
            y0 = max((y0, 0))
            for i in range(y0, y1):
                for j in range(x0, x1):
                    udg = udg_array[i][j]
                    sprite_udg = self.get_sprite_udg(state, None, udg.y - char_y, udg.x - char_x)
                    for k in range(8):
                        udg.data[k] = (udg.data[k] | sprite_udg.data[k]) & sprite_udg.mask[k]

    def _get_char_buf_descs(self):
        char_buf_descs = []
        for byte_nums, descs in self.get_sections('CharBuf', True):
            char_buf_descs.append((byte_nums, descs))
        return char_buf_descs

    def as_img(self, cwd, num, scale=2, mask=1, attr=120, udg_page=None, fname_suffix=''):
        mask_infix = 'm' if mask else 'u'
        udg_page_infix = udg_page if udg_page is not None else ''
        snapshot_name = self.get_snapshot_name()
        snapshot_infix = '_{0}'.format(snapshot_name) if snapshot_name else ''
        fname = 'as{0:03d}_{1}x{2}{3}{4}{5}{6}'.format(num & 255, attr, scale, mask_infix, udg_page_infix, snapshot_infix, fname_suffix)
        asimg_path = self.image_path(fname, 'AnimatoryStateImagePath')
        if self.need_image(asimg_path):
            self.write_image(asimg_path, self.build_sprite(num, attr, udg_page), scale=scale, mask=mask)
        return self.img_element(cwd, asimg_path, "Animatory state {0}".format(num & 255))

    def play_area(self, cwd, fname, x, y, w=1, h=1, scale=2, show_chars=0, show_x=0):
        img_path = self.image_path(fname, 'PlayAreaImagePath')
        if self.need_image(img_path):
            self.write_image(img_path, self.get_skool_udgs(x, y, w, h, show_chars, show_x), scale=scale)
        return self.img_element(cwd, img_path)

    def write_playarea_tile_table(self, cwd, tiles, id_prefix='s'):
        lines = []
        lines.append('<table class="tiles">\n')
        for row in tiles:
            lines.append('<tr>\n')
            for tile in row:
                lines.append('<td>')
                if tile:
                    bubble_id = '{0}{1.x}x{1.y}'.format(id_prefix, tile)
                    title = '(x,y)=({0.x},{0.y})'.format(tile)
                    img_fname = '{0:x}{1:x}.{2}'.format(tile.ref + 256 * tile.udg_page, tile.attr, self.default_image_format)
                    img_path = join(cwd, img_fname)
                    if self.need_image(img_path):
                        self.write_image(img_path, [[tile]], scale=4)
                    img = '<img alt="{0.x},{0.y}" src="{1}"/>'.format(tile, img_fname)
                    lines.append('<div class="T" onclick="s(\'{0}\')">{1}</div>'.format(bubble_id, img))
                    lines += self._get_bubble(bubble_id, img, title, '<br/>'.join(self.get_playarea_tile_text(tile)))
                lines.append('</td>\n')
            lines.append('</tr>\n')
        lines.append('</table>\n')
        return lines

    def write_mutable_tile_tables(self, cwd):
        return []

    def play_area_tiles(self, cwd, w, h):
        lines = []
        lines += self.write_playarea_tile_table(cwd, self.get_skool_udgs(0, 0, w, h))
        lines += self.write_mutable_tile_tables(cwd)
        return ''.join(lines)

    def _animatory_state_row(self, cwd, rownum):
        lines = ['<tr>']
        for state in (rownum, rownum + 128):
            lines.append('<td><a name="{0}"></a>{1}</td>'.format(state, self.as_img(cwd, state)))
        lines.append('<td>{0}, {1}: {2}</td>'.format(rownum, rownum + 128, self.as_descs.get(rownum, '-')))
        lines.append('</tr>')
        return lines

    def animatory_states(self, cwd):
        lines = ['<table class="animatoryStates">']
        for n in range(128):
            lines += self._animatory_state_row(cwd, n)
        lines.append('</table>')
        return '\n'.join(lines)

    def _get_animatory_state_tiles_row(self, n):
        states = ((n, None),)
        states_desc = '{0}: {1}'.format(n, self.as_descs[n])
        return [(states, states_desc)]

    def _get_bubble(self, bubble_id, img, title, text):
        return [
            '<div id="{0}" class="B P" onclick="h(this)">'.format(bubble_id),
            '<div class="H">{0}</div>'.format(title),
            '<table><tr><td>{0}</td><td>{1}</td></tr></table>'.format(img, text),
            '</div>'
        ]

    def _get_sprite_tile_img_fname(self, tile, state):
        return '{0:x}.{1}'.format(tile.ref + 256 * tile.udg_page, self.default_image_format)

    def astiles(self, cwd):
        lines = []
        attr = 120
        lines.append('<table class="asTiles">\n')
        for n in range(128):
            row = []
            for state_specs, states_desc in self._get_animatory_state_tiles_row(n):
                lines.append('<tr>\n')
                for state, udg_page in state_specs:
                    lines.append('<td>\n')
                    lines.append('<table class="A">\n')
                    sprite = self.build_sprite(state, attr, udg_page)
                    num_rows = len(sprite)
                    for row_num in range(num_rows):
                        lines.append('<tr>\n')
                        row = sprite[row_num]
                        for col_num in range(len(row)):
                            tile = row[col_num]
                            bubble_id_suffix = '{0:02x}'.format(udg_page) if udg_page is not None else ''
                            bubble_id = 'B{0:02x}{1:x}{2}'.format(state, row_num + num_rows * col_num, bubble_id_suffix)
                            title = '{0}: row {1}, column {2}'.format(state, row_num, col_num)
                            img_fname = self._get_sprite_tile_img_fname(tile, state)
                            img_path = join(cwd, img_fname)
                            if self.need_image(img_path):
                                self.write_image(img_path, [[tile]], scale=4, mask=1)
                            img = '<img alt="{0}:{1},{2}" src="{3}"/>'.format(state, row_num, col_num, img_fname)
                            text = ['<i>lsb</i>={0}, <i>ref_page</i>={1}'.format(tile.ref_addr % 256, tile.ref_addr // 256)]
                            text.append('<i>ref_addr</i>=(<i>lsb</i>,<i>ref_page</i>)={0}'.format(tile.ref_addr))
                            if tile.ref == 0:
                                text.append('<i>ref</i>={0}'.format(tile.ref))
                            else:
                                text.append('<i>ref</i>={0}, <i>udg_page</i>={1}'.format(tile.ref, tile.udg_page))
                                text.append('<i>udg_addr</i>=(<i>ref</i>,<i>udg_page</i>)={0}'.format(tile.ref + 256 * tile.udg_page))
                            lines.append('<td>')
                            lines.append('<div class="T" onclick="s(\'{0}\')">{1}</div>'.format(bubble_id, img))
                            lines += self._get_bubble(bubble_id, img, title, '<br/>'.join(text))
                            lines.append('</td>\n')
                        lines.append('</tr>\n')
                    lines.append('</table>\n')
                    lines.append('</td>\n')
                lines.append('<td>{0}</td>\n'.format(states_desc))
                lines.append('</tr>\n')
        lines.append('</table>\n')
        return ''.join(lines)

    def _get_pt_link(self, cwd, char_num):
        pt_address = self.min_lesson + 256 * char_num
        pt_file = FileInfo.asm_relpath(cwd, pt_address, self.get_code_path(MAIN_CODE_ID))
        return '<a class="link" href="{0}">{1}</a>'.format(pt_file, char_num)

    def lesson_index(self, cwd):
        lines = ['<table class="lesson">']
        lines.append('<tr>')
        lines.append('<th>ID</th>')
        lines.append('<th>Lesson</th>')
        lines.append('</tr>')
        for i in range(self.min_lesson, self.max_lesson + 1):
            lesson_link = self.expand('#LESSON{0}'.format(i), cwd)
            lines.append('<tr>')
            lines.append('<td class="lessonType">{0}</td>'.format(lesson_link))
            lines.append('<td class="lesson">{0}</td>'.format(self.lesson_descs[i]))
            lines.append('</tr>')
        lines.append('</table>')
        return '\n'.join(lines)

    def _lesson_prev_next(self, prev_link, next_link, up_link):
        prev_text = next_text = up_text = ''
        if prev_link:
            prev_text = 'Prev: {0}'.format(prev_link)
        if next_link:
            next_text = 'Next: {0}'.format(next_link)
        if up_link:
            up_text = 'Up: {0}'.format(up_link)
        lines = ['<table class="asm-navigation">', '<tr>']
        for td_class, text in (('prev', prev_text), ('up', up_text), ('next', next_text)):
            if text:
                lines.append('<td class="{0}">{1}</td>'.format(td_class, text))
            else:
                lines.append('<td class="{0}"/>'.format(td_class))
        lines.append('</tr>')
        lines.append('</table>')
        return lines

    def lesson(self, cwd, lesson):
        lines = []
        index_path = FileInfo.relpath(cwd, self.paths['LessonIndex'])

        prev_link = None if lesson == self.min_lesson else self.expand('#LESSON{0}'.format(lesson - 1), cwd)
        up_link = '<a class="link" href="{0}">Index</a>'.format(index_path)
        next_link = None if lesson == self.max_lesson else self.expand('#LESSON{0}'.format(lesson + 1), cwd)
        prev_up_next = self._lesson_prev_next(prev_link, next_link, up_link)

        lines += prev_up_next
        lines.append('<div class="description">Lesson {0}: {1}</div>'.format(lesson, self.lesson_descs[lesson]))
        lines.append('<table class="lesson">')
        lines.append('<tr>')
        lines.append('<th>Character</th>')
        lines.append('<th>Address</th>')
        lines.append('<th>Command list</th>')
        lines.append('</tr>')
        for c in range(self.min_character, self.max_character + 1):
            lesson_address = lesson + 256 * c
            tap = self.snapshot[lesson_address]
            pt_link = self._get_pt_link(cwd, c)
            char_desc = self.characters[c]
            tap_address = self.tap_addresses[tap]
            tap_link = self.expand('#R{0}({1})'.format(tap_address, tap), cwd)
            tap_desc = self.tap_descs[tap]
            lines.append('<tr>')
            lines.append('<td class="lessonCharacter">{0} - {1}</td>'.format(pt_link, char_desc))
            lines.append('<td class="lessonAddress">{0}</td>'.format(lesson_address))
            lines.append('<td class="lesson">{0} - {1}</td>'.format(tap_link, tap_desc))
            lines.append('</tr>')
        lines.append('</table>')
        lines += prev_up_next

        return '\n'.join(lines)

    def personal_timetables(self, cwd):
        lines = ['<table class="timetable">']
        lines.append('<tr>')
        lines.append('<th>No.</th>')
        lines.append('<th>Character</th>')
        lines.append('</tr>')
        for c in range(self.min_character, self.max_character + 1):
            lines.append('<tr>')
            lines.append('<td class="ptCharNum">{0}</td>'.format(self._get_pt_link(cwd, c)))
            lines.append('<td class="ptCharName">{0}</td>'.format(self.characters[c]))
            lines.append('</tr>')
        lines.append('</table>')
        return '\n'.join(lines)

    def _characters_table(self, cwd, char_buffer_base_byte):
        characters = self.characters
        first_char_num = min(characters.keys())
        cols = 3

        lines = ['<table class="cbuffer">']
        lines.append('<tr>')
        for c in range(cols):
            lines.append('<th>Page</th>')
            lines.append('<th>Address</th>')
            lines.append('<th>Character</th>')
            if c < cols - 1:
                lines.append('<td class="transparent"/>')
        lines.append('</tr>')

        rows = (cols - 1  + len(characters)) // cols
        for r in range(rows):
            lines.append('<tr>')
            for c in range(cols):
                char_index = r + c * rows
                if char_index < len(characters):
                    char_num = first_char_num + char_index
                    lines.append('<td class="cbufPage">{0}</td>'.format(char_num))
                    buf_addr = char_buffer_base_byte + char_num * 256
                    entry_link = self.expand('#R{0}'.format(buf_addr), cwd)
                    lines.append('<td class="cbufAddress">{0}</td>'.format(entry_link))
                    lines.append('<td class="cbufCharacter">{0}</td>'.format(characters[char_num]))
                    if c < cols - 1:
                        lines.append('<td class="transparent"/>')
            lines.append('</tr>')

        lines.append('</table>')
        return lines

    def character_buffers(self, cwd, char_buffer_base_byte):
        lines = self._characters_table(cwd, char_buffer_base_byte)
        lines.append('<p>The bytes in each character buffer are used as follows:</p>')
        lines.append('<table class="cbuffer">')
        lines.append('<tr>')
        lines.append('<th>Byte(s)</th>')
        lines.append('<th>Contents</th>')
        lines.append('</tr>')
        for byte_nums, entries in self.char_buf_descs:
            rows = len(entries)
            lines.append('<tr>')
            rowspan_attr = '' if rows == 1 else ' rowspan="{0}"'.format(rows)
            lines.append('<td class="cbufBytes"{0}>{1}</td>'.format(rowspan_attr, byte_nums))
            lines.append('<td class="cbufByteDesc">\n{0}\n</td>'.format(self.expand(entries[0], cwd)))
            lines.append('</tr>')
            for i in range(1, rows):
                lines.append('<tr>')
                lines.append('<td class="cbufByteDesc">\n{0}\n</td>'.format(self.expand(entries[i], cwd)))
                lines.append('</tr>')
        lines.append('</table>')
        return '\n'.join(lines)

    def expand_as(self, text, index, cwd):
        # #AS[state][(link text)]
        end, state, link_text = parse_params(text, index)
        as_file = FileInfo.relpath(cwd, self.paths['AnimatoryStates'])
        anchor = '#{0}'.format(state) if state else ''
        link_text = link_text or state
        link = '<a class="link" href="{0}{1}">{2}</a>'.format(as_file, anchor, link_text)
        return end, link

    def expand_lesson(self, text, index, cwd):
        # #LESSONnum
        end, lesson = parse_ints(text, index, 1)
        page_id = 'Lesson{0}'.format(lesson)
        href = FileInfo.relpath(cwd, self.paths[page_id])
        link = '<a class="link" href="{0}">{1}</a>'.format(href, lesson)
        return end, link

class MicrosphereAsmWriter(AsmWriter):
    def expand_as(self, text, index):
        # #AS[state][(link text)]
        end, state, link_text = parse_params(text, index)
        return end, link_text or state

    def expand_lesson(self, text, index):
        # #LESSONnum
        end, lesson = parse_ints(text, index, 1)
        return end, str(lesson)

class Udg(BaseUdg):
    def __init__(self, attr, data, mask=None, attr_addr=None, q_addr=None, ref_addr=None, ref=None, udg_page=None, x=None, y=None, fg_udg=None):
        BaseUdg.__init__(self, attr, data, mask)
        self.attr_addr = attr_addr
        self.q_addr = q_addr
        self.ref_addr = ref_addr
        # We store the UDG reference now in case the snapshot changes before
        # the reference is looked up via self.snapshot[udg.ref_addr] (e.g.
        # because a saved snapshot is restored after
        # BackToSkoolHtmlWriter.alter_skool_udgs() has been called)
        self.ref = ref
        self.udg_page = udg_page
        self.x = x
        self.y = y
        self.fg_udg = fg_udg
